@@ -1,4 +1,4 @@
-# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
+# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
 #
 # Copyright 2002 Ben Escoto <ben@emerose.org>
 # Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
@@ -19,45 +19,40 @@
 # along with duplicity; if not, write to the Free Software Foundation,
 # Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
-u"""
+"""
 Provides a common interface to all backends and certain sevices
 intended to be used by the backends themselves.
 """
 
-from future import standard_library
-standard_library.install_aliases()
-from builtins import str
-from builtins import range
-from builtins import object
-
+from datetime import datetime
 import errno
+import getpass
+import multiprocessing
 import os
+import re
 import sys
 import time
-import re
-import getpass
-import re
-import urllib.request  # pylint: disable=import-error
-import urllib.parse  # pylint: disable=import-error
-import urllib.error  # pylint: disable=import-error
+import traceback
+from typing import Tuple
+import urllib.error
+import urllib.parse
+import urllib.request
+from duplicity import errors
 
+import duplicity.backends
+from duplicity import config
 from duplicity import dup_temp
 from duplicity import file_naming
-from duplicity import config
 from duplicity import log
 from duplicity import path
 from duplicity import util
-
-from duplicity.util import exception_traceback
-
 from duplicity.errors import BackendException
-from duplicity.errors import FatalBackendException
-from duplicity.errors import TemporaryLoadException
 from duplicity.errors import ConflictingScheme
+from duplicity.errors import FatalBackendException
 from duplicity.errors import InvalidBackendURL
+from duplicity.errors import TemporaryLoadException
 from duplicity.errors import UnsupportedBackendScheme
-
-import duplicity.backends
+from duplicity.util import exception_traceback
 
 _backends = {}
 _backend_prefixes = {}
@@ -78,7 +73,7 @@ uses_netloc = []
 
 
 def import_backends():
-    u"""
+    """
     Import files in the duplicity/backends directory where
     the filename ends in 'backend.py' and ignore the rest.
 
@@ -86,26 +81,25 @@ def import_backends():
     @return: void
     """
     path = duplicity.backends.__path__[0]
-    assert path.endswith(u"duplicity/backends"), duplicity.backends.__path__
+    assert path.endswith("duplicity/backends"), duplicity.backends.__path__
 
-    files = os.listdir(path)
-    files.sort()
+    files = sorted(os.listdir(path))
     for fn in files:
-        if fn.endswith(u"backend.py"):
+        if fn.endswith("backend.py"):
             fn = fn[:-3]
-            imp = u"duplicity.backends.%s" % (fn,)
+            imp = f"duplicity.backends.{fn}"
             try:
                 __import__(imp)
-                res = u"Succeeded"
+                res = "Succeeded"
             except Exception:
-                res = u"Failed: " + str(sys.exc_info()[1])
-            log.Log(_(u"Import of %s %s") % (imp, res), log.INFO)
+                res = f"Failed: {str(sys.exc_info()[1])}"
+            log.Log(_("Import of %s %s") % (imp, res), log.INFO)
         else:
             continue
 
 
 def register_backend(scheme, backend_factory):
-    u"""
+    """
     Register a given backend factory responsible for URL:s with the
     given scheme.
 
@@ -120,18 +114,16 @@ def register_backend(scheme, backend_factory):
     """
     global _backends
 
-    assert callable(backend_factory), u"backend factory must be callable"
+    assert callable(backend_factory), "backend factory must be callable"
 
     if scheme in _backends:
-        raise ConflictingScheme(u"the scheme %s already has a backend "
-                                u"associated with it"
-                                u"" % (scheme,))
+        raise ConflictingScheme(f"the scheme {scheme} already has a backend associated with it")
 
     _backends[scheme] = backend_factory
 
 
 def register_backend_prefix(scheme, backend_factory):
-    u"""
+    """
     Register a given backend factory responsible for URL:s with the
     given scheme prefix.
 
@@ -146,25 +138,23 @@ def register_backend_prefix(scheme, backend_factory):
     """
     global _backend_prefixes
 
-    assert callable(backend_factory), u"backend factory must be callable"
+    assert callable(backend_factory), "backend factory must be callable"
 
     if scheme in _backend_prefixes:
-        raise ConflictingScheme(u"the prefix %s already has a backend "
-                                u"associated with it"
-                                u"" % (scheme,))
+        raise ConflictingScheme(f"the prefix {scheme} already has a backend associated with it")
 
     _backend_prefixes[scheme] = backend_factory
 
 
 def strip_prefix(url_string, prefix_scheme):
-    u"""
+    """
     strip the prefix from a string e.g. par2+ftp://... -> ftp://...
     """
-    return re.sub(r'(?i)^' + re.escape(prefix_scheme) + r'\+', r'', url_string)
+    return re.sub(f"(?i)^{re.escape(prefix_scheme)}\\+", r"", url_string)
 
 
 def is_backend_url(url_string):
-    u"""
+    """
     @return Whether the given string looks like a backend URL.
     """
     pu = ParsedUrl(url_string)
@@ -177,7 +167,7 @@ def is_backend_url(url_string):
 
 
 def get_backend_object(url_string):
-    u"""
+    """
     Find the right backend class instance for the given URL, or return None
     if the given string looks like a local path rather than a URL.
 
@@ -189,12 +179,12 @@ def get_backend_object(url_string):
     global _backends, _backend_prefixes
 
     pu = ParsedUrl(url_string)
-    assert pu.scheme, u"should be a backend url according to is_backend_url"
+    assert pu.scheme, "should be a backend url according to is_backend_url"
 
     factory = None
 
     for prefix in _backend_prefixes:
-        if url_string.startswith(prefix + u'+'):
+        if url_string.startswith(f"{prefix}+"):
             factory = _backend_prefixes[prefix]
             pu = ParsedUrl(strip_prefix(url_string, prefix))
             break
@@ -208,18 +198,16 @@ def get_backend_object(url_string):
     try:
         return factory(pu)
     except ImportError:
-        raise BackendException(_(u"Could not initialize backend: %s") % str(sys.exc_info()[1]))
+        raise BackendException(_("Could not initialize backend: %s") % str(sys.exc_info()[1]))
 
 
 def get_backend(url_string):
-    u"""
+    """
     Instantiate a backend suitable for the given URL, or return None
     if the given string looks like a local path rather than a URL.
 
     Raise InvalidBackendURL if the URL is not a valid URL.
     """
-    if config.use_gio:
-        url_string = u'gio+' + url_string
     obj = get_backend_object(url_string)
     if obj:
         obj = BackendWrapper(obj)
@@ -227,7 +215,7 @@ def get_backend(url_string):
 
 
 class ParsedUrl(object):
-    u"""
+    """
     Parse the given URL as a duplicity backend URL.
 
     Returns the data of a parsed URL with the same names as that of
@@ -237,6 +225,7 @@ class ParsedUrl(object):
 
     Raise InvalidBackendURL on invalid URL's
     """
+
     def __init__(self, url_string):
         self.url_string = url_string
 
@@ -252,29 +241,29 @@ class ParsedUrl(object):
         try:
             pu = urllib.parse.urlparse(url_string)
         except Exception:
-            raise InvalidBackendURL(u"Syntax error in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error in: {url_string}")
 
         try:
             self.scheme = pu.scheme
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (scheme) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (scheme) in: {url_string}")
 
         try:
             self.netloc = pu.netloc
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (netloc) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (netloc) in: {url_string}")
 
         try:
             self.path = pu.path
             if self.path:
                 self.path = urllib.parse.unquote(self.path)
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (path) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (path) in: {url_string}")
 
         try:
             self.username = pu.username
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (username) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (username) in: {url_string}")
         if self.username:
             self.username = urllib.parse.unquote(pu.username)
         else:
@@ -283,7 +272,7 @@ class ParsedUrl(object):
         try:
             self.password = pu.password
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (password) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (password) in: {url_string}")
         if self.password:
             self.password = urllib.parse.unquote(self.password)
         else:
@@ -292,12 +281,12 @@ class ParsedUrl(object):
         try:
             self.hostname = pu.hostname
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (hostname) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (hostname) in: {url_string}")
 
         try:
             self.query = pu.query
         except Exception:
-            raise InvalidBackendURL(u"Syntax error (query) in: %s" % url_string)
+            raise InvalidBackendURL(f"Syntax error (query) in: {url_string}")
         if self.query:
             self.query_args = urllib.parse.parse_qs(self.query)
         else:
@@ -308,26 +297,26 @@ class ParsedUrl(object):
         self.port = None
         try:
             self.port = pu.port
-        except Exception:  # not raised in python2.7, just returns None
-            # TODO: remove after dropping python 2.7 support
-            if self.scheme in [u'rclone']:
+        except Exception:  # just returns None
+            if self.scheme in ["rclone"]:
                 pass
             # old style rsync://host::[/]dest, are still valid, though they contain no port
-            elif not (u'rsync' in self.scheme and re.search(u'::[^:]*$', self.url_string)):
-                raise InvalidBackendURL(u"Syntax error (port) in: %s A%s B%s C%s" %
-                                        (url_string, (u'rsync' in self.scheme),
-                                         re.search(u'::[^:]+$', self.netloc), self.netloc))
+            elif not ("rsync" in self.scheme and re.search("::[^:]*$", self.url_string)):
+                raise InvalidBackendURL(
+                    f"Syntax error (port) in: {url_string} A{'rsync' in self.scheme} "
+                    f"B{re.search('::[^:]+$', self.netloc)} C{self.netloc}"
+                )
 
         # Our URL system uses two slashes more than urlparse's does when using
         # non-netloc URLs.  And we want to make sure that if urlparse assuming
         # a netloc where we don't want one, that we correct it.
         if self.scheme not in uses_netloc:
             if self.netloc:
-                self.path = u'//' + self.netloc + self.path
-                self.netloc = u''
+                self.path = f"//{self.netloc}{self.path}"
+                self.netloc = ""
                 self.hostname = None
-            elif not self.path.startswith(u'//') and self.path.startswith(u'/'):
-                self.path = u'//' + self.path
+            elif not self.path.startswith("//") and self.path.startswith("/"):
+                self.path = f"//{self.path}"
 
         # This happens for implicit local paths.
         if not self.scheme:
@@ -335,15 +324,15 @@ class ParsedUrl(object):
 
         # Our backends do not handle implicit hosts.
         if self.scheme in uses_netloc and not self.hostname:
-            raise InvalidBackendURL(u"Missing hostname in a backend URL which "
-                                    u"requires an explicit hostname: %s"
-                                    u"" % (url_string))
+            raise InvalidBackendURL(
+                f"Missing hostname in a backend URL which requires an " f"explicit hostname: {url_string}"
+            )
 
         # Our backends do not handle implicit relative paths.
-        if self.scheme not in uses_netloc and not self.path.startswith(u'//'):
-            raise InvalidBackendURL(u"missing // - relative paths not supported "
-                                    u"for scheme %s: %s"
-                                    u"" % (self.scheme, url_string))
+        if self.scheme not in uses_netloc and not self.path.startswith("//"):
+            raise InvalidBackendURL(
+                f"missing // - relative paths not supported for " f"scheme {self.scheme}: {url_string}"
+            )
 
     def geturl(self):
         return self.url_string
@@ -353,18 +342,18 @@ class ParsedUrl(object):
 
 
 def strip_auth_from_url(parsed_url):
-    u"""Return a URL from a urlparse object without a username or password."""
+    """Return a URL from a urlparse object without a username or password."""
 
-    clean_url = re.sub(u'^([^:/]+://)(.*@)?(.*)', r'\1\3', parsed_url.geturl())
+    clean_url = re.sub("^([^:/]+://)(.*@)?(.*)", r"\1\3", parsed_url.geturl())
     return clean_url
 
 
 def _get_code_from_exception(backend, operation, e):
     if isinstance(e, BackendException) and e.code != log.ErrorCode.backend_error:
         return e.code
-    elif hasattr(backend, u'_error_code'):
+    elif hasattr(backend, "_error_code"):
         return backend._error_code(operation, e) or log.ErrorCode.backend_error
-    elif hasattr(e, u'errno'):
+    elif hasattr(e, "errno"):
         # A few backends return such errors (local, paramiko, etc)
         if e.errno == errno.EACCES:
             return log.ErrorCode.backend_permission_denied
@@ -400,8 +389,7 @@ def retry(operation, fatal=True):
                         return errors_default
                     else:
                         # retry on anything else
-                        log.Debug(_(u"Backtrace of previous error: %s")
-                                  % exception_traceback())
+                        log.Debug(_("Backtrace of previous error: %s") % exception_traceback())
                         at_end = n == config.num_retries
                         code = _get_code_from_exception(self.backend, operation, e)
                         if code == log.ErrorCode.backend_not_found:
@@ -409,43 +397,63 @@ def retry(operation, fatal=True):
                             # no need to retry.
                             at_end = True
                         if at_end and fatal:
+
                             def make_filename(f):
                                 if isinstance(f, path.ROPath):
                                     return util.escape(f.uc_name)
                                 else:
                                     return util.escape(f)
-                            extra = u' '.join([operation] + [make_filename(x) for x in args
-                                                             if (x and isinstance(x, str))])
-                            log.FatalError(_(u"Giving up after %s attempts. %s: %s")
-                                           % (n, e.__class__.__name__,
-                                              util.uexc(e)), code=code, extra=extra)
+
+                            extra = " ".join(
+                                [operation] + [make_filename(x) for x in args if (x and isinstance(x, str))]
+                            )
+                            if multiprocessing.parent_process():
+                                # running as a child process we need to raise an exception to signal an issue
+                                log.Error(
+                                    _("Giving up after %s attempts. %s: %s. (for trace back: set log level DEBUG)")
+                                    % (n, e.__class__.__name__, util.uexc(e)),
+                                    code=code,
+                                    extra=extra,
+                                )
+                                e.code = code
+                                raise
+                            else:
+                                log.FatalError(
+                                    _("Giving up after %s attempts. %s: %s") % (n, e.__class__.__name__, util.uexc(e)),
+                                    code=code,
+                                    extra=extra,
+                                )
                         else:
-                            log.Warn(_(u"Attempt of %s Nr. %s failed. %s: %s")
-                                     % (fn.__name__, n, e.__class__.__name__, util.uexc(e)))
+                            log.Warn(
+                                _("Attempt of %s Nr. %s failed. %s: %s")
+                                % (fn.__name__, n, e.__class__.__name__, util.uexc(e))
+                            )
                         if not at_end:
                             if isinstance(e, TemporaryLoadException):
                                 time.sleep(3 * config.backend_retry_delay)  # wait longer before trying again
                             else:
                                 time.sleep(config.backend_retry_delay)  # wait a bit before trying again
-                            if hasattr(self.backend, u'_retry_cleanup'):
+                            if hasattr(self.backend, "_retry_cleanup"):
                                 self.backend._retry_cleanup()
 
         return inner_retry
+
     return outer_retry
 
 
 class Backend(object):
-    u"""
+    """
     See README in backends directory for information on how to write a backend.
     """
+
     def __init__(self, parsed_url):
         self.parsed_url = parsed_url
 
-    u""" use getpass by default, inherited backends may overwrite this behaviour """
+    """ use getpass by default, inherited backends may overwrite this behaviour """
     use_getpass = True
 
     def get_password(self):
-        u"""
+        """
         Return a password for authentication purposes. The password
         will be obtained from the backend URL, the environment, by
         asking the user, or by some other method. When applicable, the
@@ -455,18 +463,17 @@ class Backend(object):
             return self.parsed_url.password
 
         try:
-            password = os.environ[u'FTP_PASSWORD']
+            password = os.environ["FTP_PASSWORD"]
         except KeyError:
             if self.use_getpass:
-                password = getpass.getpass(u"Password for '%s@%s': " %
-                                           (self.parsed_url.username, self.parsed_url.hostname))
-                os.environ[u'FTP_PASSWORD'] = password
+                password = getpass.getpass(f"Password for '{self.parsed_url.username}@{self.parsed_url.hostname}': ")
+                os.environ["FTP_PASSWORD"] = password
             else:
                 password = None
         return password
 
     def munge_password(self, commandline):
-        u"""
+        """
         Remove password from commandline by substituting the password
         found in the URL, if any, with a generic place-holder.
 
@@ -475,17 +482,20 @@ class Backend(object):
         the ':password@' may be substituted.
         """
         if self.parsed_url.password:
-            return re.sub(r'(:([^\s:/@]+)@([^\s@]+))', r':*****@\3', commandline)
+            return re.sub(r"(:([^\s:/@]+)@([^\s@]+))", r":*****@\3", commandline)
         else:
             return commandline
 
     def __subprocess_popen(self, args):
-        u"""
+        """
         For internal use.
         Execute the given command line, interpreted as a shell command.
         Returns int Exitcode, string StdOut, string StdErr
         """
-        from subprocess import Popen, PIPE
+        from subprocess import (
+            Popen,
+            PIPE,
+        )
 
         args[0] = util.which(args[0])
         p = Popen(args, stdout=PIPE, stderr=PIPE, universal_newlines=True)
@@ -493,12 +503,12 @@ class Backend(object):
 
         return p.returncode, stdout, stderr
 
-    u""" a dictionary for breaking exceptions, syntax is
+    """ a dictionary for breaking exceptions, syntax is
         { 'command' : [ code1, code2 ], ... } see ftpbackend for an example """
     popen_breaks = {}
 
     def subprocess_popen(self, commandline):
-        u"""
+        """
         Execute the given command line with error check.
         Returns int Exitcode, string StdOut, string StdErr
 
@@ -507,30 +517,31 @@ class Backend(object):
         import shlex
 
         if isinstance(commandline, (list, tuple)):
-            logstr = u' '.join(commandline)
+            logstr = " ".join(commandline)
             args = commandline
         else:
             logstr = commandline
             args = shlex.split(commandline)
 
         logstr = self.munge_password(logstr)
-        log.Info(_(u"Reading results of '%s'") % logstr)
+        log.Info(_("Reading results of '%s'") % logstr)
 
         result, stdout, stderr = self.__subprocess_popen(args)
         if result != 0:
             try:
                 ignores = self.popen_breaks[args[0]]
                 ignores.index(result)
-                u""" ignore a predefined set of error codes """
-                return 0, u'', u''
+                """ ignore a predefined set of error codes """
+                return 0, "", ""
             except (KeyError, ValueError):
-                raise BackendException(u"Error running '%s': returned %d, with output:\n%s" %
-                                       (logstr, result, stdout + u'\n' + stderr + u'\n'))
+                raise BackendException(
+                    "Error running '%s': returned %d, with output:\n%s" % (logstr, result, f"{stdout}\n{stderr}\n")
+                )
         return result, stdout, stderr
 
 
 class BackendWrapper(object):
-    u"""
+    """
     Represents a generic duplicity backend, capable of storing and
     retrieving files.
     """
@@ -539,15 +550,15 @@ class BackendWrapper(object):
         self.backend = backend
 
     def __do_put(self, source_path, remote_filename):
-        if hasattr(self.backend, u'_put'):
-            log.Info(_(u"Writing %s") % util.fsdecode(remote_filename))
+        if hasattr(self.backend, "_put"):
+            log.Info(_("Writing %s") % os.fsdecode(remote_filename))
             self.backend._put(source_path, remote_filename)
         else:
             raise NotImplementedError()
 
-    @retry(u'put', fatal=True)
+    @retry("put", fatal=True)
     def put(self, source_path, remote_filename=None):
-        u"""
+        """
         Transfer source_path (Path object) to remote_filename (string)
 
         If remote_filename is None, get the filename from the last
@@ -555,11 +566,25 @@ class BackendWrapper(object):
         """
         if not remote_filename:
             remote_filename = source_path.get_filename()
+        if f"vol{config.put_fail_volume}.difftar" in os.fsdecode(remote_filename):
+            raise FileNotFoundError(f"Forced error for testing at volume {config.put_fail_volume}")
         self.__do_put(source_path, remote_filename)
 
-    @retry(u'move', fatal=True)
+    def put_validated(self, source_path, remote_filename):
+        self.put(source_path, remote_filename)
+        size = source_path.getsize()
+        res, msg = self.validate(remote_filename, size, source_path)
+        if res:
+            return size
+        else:
+            raise errors.BackendException(
+                f"Remotefile {remote_filename}: Transfer not successful. Validation failed: {msg}",
+                code=log.ErrorCode.backend_validation_failed,
+            )
+
+    @retry("move", fatal=True)
     def move(self, source_path, remote_filename=None):
-        u"""
+        """
         Move source_path (Path object) to remote_filename (string)
 
         Same as put(), but unlinks source_path in the process.  This allows the
@@ -567,84 +592,120 @@ class BackendWrapper(object):
         """
         if not remote_filename:
             remote_filename = source_path.get_filename()
-        if hasattr(self.backend, u'_move'):
+        if hasattr(self.backend, "_move"):
             if self.backend._move(source_path, remote_filename) is not False:
                 source_path.setdata()
                 return
         self.__do_put(source_path, remote_filename)
         source_path.delete()
 
-    @retry(u'get', fatal=True)
+    @retry("get", fatal=True)
     def get(self, remote_filename, local_path):
-        u"""Retrieve remote_filename and place in local_path"""
-        if hasattr(self.backend, u'_get'):
+        """Retrieve remote_filename and place in local_path"""
+        if hasattr(self.backend, "_get"):
             self.backend._get(remote_filename, local_path)
             local_path.setdata()
             if not local_path.exists():
-                raise BackendException(_(u"File %s not found locally after get "
-                                         u"from backend") % local_path.uc_name)
+                raise BackendException(_("File %s not found locally after get " "from backend") % local_path.uc_name)
         else:
             raise NotImplementedError()
 
-    @retry(u'list', fatal=True)
+    @retry("list", fatal=True)
     def list(self):
-        u"""
+        """
         Return list of filenames (byte strings) present in backend
         """
+
         def tobytes(filename):
-            u"Convert a (maybe unicode) filename to bytes"
+            """Convert a (maybe unicode) filename to bytes"""
             if isinstance(filename, str):
                 # There shouldn't be any encoding errors for files we care
                 # about, since duplicity filenames are ascii.  But user files
                 # may be in the same directory.  So just replace characters.
-                return util.fsencode(filename)
+                return os.fsencode(filename)
             else:
                 return filename
 
-        if hasattr(self.backend, u'_list'):
+        if hasattr(self.backend, "_list"):
             # Make sure that duplicity internals only ever see byte strings
             # for filenames, no matter what the backend thinks it is talking.
             return [tobytes(x) for x in self.backend._list()]
         else:
             raise NotImplementedError()
 
+    def validate(self, remote_filename, expected_size, source_path=None) -> Tuple[bool, str]:
+        """
+        validates a file transferred to the backend by comparing the size.
+        source_path is optional to allow the backend do further validation, e.g. by
+        calulating a hash or so.
+        Returns: return a tuple of (result, reason)
+            - result: bool: True if file validation successful otherwise False
+            - reason: str: If False a reason why validation failed; True: emtpy str or description
+        """
+        msg = "Validation fail for unknow reaons."
+        if hasattr(self.backend, "_validate"):
+            return self.backend._validate(remote_filename, expected_size, source_path=source_path)
+        else:
+            # as some backends take some time refresh file attributes after upload
+            # retry config.num_retries times.
+            for attempt in range(0, config.num_retries + 1):
+                info = self.query_info([remote_filename])[remote_filename]
+                size = info["size"]
+                if size is None:
+                    log.Debug(
+                        "File size can't be validated, because of missing capabilities of the backend. "
+                        "Please verify the backup separately."
+                    )
+                    return True, "Backend has no capabilities to check filesize, skip validation."
+                if size == expected_size:
+                    return True, "Validation OK, file size matches."
+                msg = _("%s Remote filesize %d for %s does not match local size %d") % (
+                    datetime.now(),
+                    size,
+                    util.escape(remote_filename),
+                    expected_size,
+                )
+                log.Notice(f"{msg}, retrying.")
+                time.sleep(2**attempt)
+        return False, f"{msg}."
+
     def pre_process_download(self, remote_filename):
-        u"""
+        """
         Manages remote access before downloading files
         (unseal data in cold storage for instance)
         """
-        if hasattr(self.backend, u'pre_process_download'):
+        if hasattr(self.backend, "pre_process_download"):
             return self.backend.pre_process_download(remote_filename)
 
     def pre_process_download_batch(self, remote_filenames):
-        u"""
+        """
         Manages remote access before downloading files
         (unseal data in cold storage for instance)
         """
-        if hasattr(self.backend, u'pre_process_download_batch'):
+        if hasattr(self.backend, "pre_process_download_batch"):
             return self.backend.pre_process_download_batch(remote_filenames)
 
     def delete(self, filename_list):
-        u"""
+        """
         Delete each filename in filename_list, in order if possible.
         """
         assert not isinstance(filename_list, bytes)
-        if hasattr(self.backend, u'_delete_list'):
+        if hasattr(self.backend, "_delete_list"):
             self._do_delete_list(filename_list)
-        elif hasattr(self.backend, u'_delete'):
+        elif hasattr(self.backend, "_delete"):
             for filename in filename_list:
                 self._do_delete(filename)
         else:
             raise NotImplementedError()
 
-    @retry(u'delete', fatal=False)
+    @retry("delete", fatal=False)
     def _do_delete_list(self, filename_list):
         while filename_list:
             sublist = filename_list[:100]
             self.backend._delete_list(sublist)
             filename_list = filename_list[100:]
 
-    @retry(u'delete', fatal=False)
+    @retry("delete", fatal=False)
     def _do_delete(self, filename):
         self.backend._delete(filename)
 
@@ -659,15 +720,15 @@ class BackendWrapper(object):
     # Returned dictionary is guaranteed to contain a metadata dictionary for
     # each filename, and all metadata are guaranteed to be present.
     def query_info(self, filename_list):
-        u"""
+        """
         Return metadata about each filename in filename_list
         """
         info = {}
-        if hasattr(self.backend, u'_query_list'):
+        if hasattr(self.backend, "_query_list"):
             info = self._do_query_list(filename_list)
             if info is None:
                 info = {}
-        elif hasattr(self.backend, u'_query'):
+        elif hasattr(self.backend, "_query"):
             for filename in filename_list:
                 info[filename] = self._do_query(filename)
 
@@ -676,39 +737,39 @@ class BackendWrapper(object):
         for filename in filename_list:
             if filename not in info or info[filename] is None:
                 info[filename] = {}
-            for metadata in [u'size']:
+            for metadata in ["size"]:
                 info[filename].setdefault(metadata, None)
 
         return info
 
-    @retry(u'query', fatal=False)
+    @retry("query", fatal=False)
     def _do_query_list(self, filename_list):
         info = self.backend._query_list(filename_list)
         if info is None:
             info = {}
         return info
 
-    @retry(u'query', fatal=False)
+    @retry("query", fatal=False)
     def _do_query(self, filename):
         try:
             return self.backend._query(filename)
         except Exception as e:
-            code = _get_code_from_exception(self.backend, u'query', e)
+            code = _get_code_from_exception(self.backend, "query", e)
             if code == log.ErrorCode.backend_not_found:
-                return {u'size': -1}
+                return {"size": -1}
             else:
                 raise e
 
     def close(self):
-        u"""
+        """
         Close the backend, releasing any resources held and
         invalidating any file objects obtained from the backend.
         """
-        if hasattr(self.backend, u'_close'):
+        if hasattr(self.backend, "_close"):
             self.backend._close()
 
     def get_fileobj_read(self, filename, parseresults=None):
-        u"""
+        """
         Return fileobject opened for reading of filename on backend
 
         The file will be downloaded first into a temp file.  When the
@@ -716,14 +777,14 @@ class BackendWrapper(object):
         """
         if not parseresults:
             parseresults = file_naming.parse(filename)
-            assert parseresults, u"Filename not correctly parsed"
+            assert parseresults, "Filename not correctly parsed"
         tdp = dup_temp.new_tempduppath(parseresults)
         self.get(filename, tdp)
         tdp.setdata()
-        return tdp.filtered_open_with_delete(u"rb")
+        return tdp.filtered_open_with_delete("rb")
 
     def get_data(self, filename, parseresults=None):
-        u"""
+        """
         Retrieve a file from backend, process it, return contents.
         """
         fin = self.get_fileobj_read(filename, parseresults)
